Gdk.Pixbuf resize_pixbuf(Gdk.Pixbuf pixbuf, Dimensions resized, Gdk.InterpType interp) { return pixbuf; }
Gdk.Pixbuf scale_pixbuf(Gdk.Pixbuf pixbuf, int scale, Gdk.InterpType interp, bool scale_up) { return pixbuf; }
Gdk.Pixbuf apply_alpha_channel(Gdk.Pixbuf pixbuf, bool strip = false) { return pixbuf; }