// stub class for photo metadata needed by the JFIF Support test
public class PhotoMetadata {
    public PhotoMetadata() { }
    public void read_from_file(File file) throws Error { }
    public void write_to_file(File file) throws Error { }
    public string? exif_hash() { return null; }
    public string? thumbnail_hash() { return null; }
}
